// Copyright (c) 2016, Shinichiro Abe. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

import 'dart:async';

import 'package:ducene/analysis.dart';
import 'package:ducene/index.dart';
import 'package:ducene/search.dart';
import 'package:ducene/store.dart';

class IndexWriterBenchmark {
  static Future<Null> main() async {
    for (int i = 0; i < 10; i++) {
      final IndexWriterBenchmark bench = new IndexWriterBenchmark();
      await bench.run();
    }
  }

  Future<Null> run() async {
    final Analyzer ws = new WhitespaceAnalyzer(); // slow new NGramAnalyzer(2);
    final IndexDirectory directory = new RAMIndexDirectory();
    final IndexWriter writer = new IndexWriter(directory);
    final int iteration = 10000;
    final Iterable<Document> docs =
        new Iterable<Document>.generate(iteration, (int i) {
      return new Document()
          .append('id', i.toString())
          .append('text', text, analyzer: ws);
    });
    final Stopwatch stopwatch = new Stopwatch();
    stopwatch.start();
    await writer.write(docs);
    print('indexed ${stopwatch.elapsed.inMilliseconds} ms');
  }
}

const String text = '''
Dart is a general-purpose programming language originally developed
by Google and later approved as a standard by Ecma (ECMA-408).[4]
It is used to build web, server and mobile applications,
and for Internet of Things (IoT) devices.[5]
It is open-source software under a BSD license.
''';

class IndexSearcherBenchmark {
  static Future<Null> main() async {
    for (int i = 0; i < 10; i++) {
      final IndexSearcherBenchmark bench = new IndexSearcherBenchmark();
      await bench.setup();
      await bench.run();
    }
  }

  Analyzer _st;
  IndexDirectory _directory;

  Future<Null> setup() async {
    _st = new StandardAnalyzer();
    _directory = new RAMIndexDirectory();
    final IndexWriter writer = new IndexWriter(_directory);
    final int iteration = 10000;
    final Iterable<Document> docs =
        new Iterable<Document>.generate(iteration, (int i) {
      return new Document()
          .append('id', i.toString())
          .append('text', text, analyzer: _st);
    });
    //Stopwatch stopwatch = new Stopwatch();
    //stopwatch.start();
    await writer.write(docs);
    //print('indexed ${stopwatch.elapsed.inMilliseconds} ms');
  }

  Future<Null> run() async {
    final IndexReader reader = await DirectoryReader.open(_directory);
    final IndexSearcher searcher = new IndexSearcher(reader);
    final BoolQuery query =
        new BoolQuery(Op.or).append('text', 'software', analyzer: _st);
    final Stopwatch stopwatch = new Stopwatch();
    stopwatch.start();
    final TopDocs hits = await searcher.search(query, 10);
    assert(10000 == hits.totalHits);
    print('searched ${stopwatch.elapsed.inMilliseconds} ms');
  }
}

Future<Null> main() async {
  await IndexWriterBenchmark.main();
  await IndexSearcherBenchmark.main();
}
